use crate::ast::Identifier;
use std::fmt::Display;
use std::range::Range;

#[derive(Clone, Debug)]
pub struct SourceDefinition {
    file_id: usize,
    range: Range<usize>,
}

impl SourceDefinition {
    pub fn from_identifier(identifier: &Identifier) -> Self {
        SourceDefinition {
            file_id: identifier.file_id,
            range: identifier.range,
        }
    }

    pub fn file_id(&self) -> usize {
        self.file_id
    }

    pub fn range(&self) -> Range<usize> {
        self.range.clone()
    }
}

#[derive(Debug, Clone)]
pub enum Symbol {
    Function(FunctionSymbol),
    Variable(VariableSymbol),
    Module(ModuleSymbol),
}

impl Symbol {
    pub fn name(&self) -> &str {
        match self {
            Symbol::Function(s) => s.declared_name.as_str(),
            Symbol::Variable(s) => s.name.as_str(),
            Symbol::Module(s) => s.declared_name.as_str(),
        }
    }

    pub fn definition(&self) -> &SourceDefinition {
        match self {
            Symbol::Function(s) => s.definition(),
            Symbol::Module(s) => s.definition(),
            Symbol::Variable(s) => s.definition(),
        }
    }
}

impl Display for Symbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        use Symbol::*;
        match self {
            Function(function_symbol) => write!(f, "{}", function_symbol),
            Variable(variable_symbol) => write!(f, "{}", variable_symbol),
            Module(module_symbol) => write!(f, "{}", module_symbol),
        }
    }
}

#[derive(Debug, Clone)]
pub struct FunctionSymbol {
    pub fqn: String,
    pub declared_name: String,
    pub is_public: bool,
    definition: SourceDefinition,
}

impl FunctionSymbol {
    pub fn new(
        fqn: &str,
        declared_name: &str,
        is_public: bool,
        identifier: &Identifier,
    ) -> FunctionSymbol {
        FunctionSymbol {
            fqn: fqn.to_string(),
            declared_name: declared_name.to_string(),
            is_public,
            definition: SourceDefinition::from_identifier(identifier),
        }
    }

    fn definition(&self) -> &SourceDefinition {
        &self.definition
    }
}

impl Display for FunctionSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "FunctionSymbol(fqn = {}, declared_name = {}, is_public = {})",
            self.fqn, self.declared_name, self.is_public
        )
    }
}

#[derive(Debug, Clone)]
pub struct VariableSymbol {
    pub name: String,
    pub is_mutable: bool,
    definition: SourceDefinition,
}

impl VariableSymbol {
    pub fn new(name: &str, is_mutable: bool, definition: SourceDefinition) -> Self {
        VariableSymbol {
            name: name.to_string(),
            is_mutable,
            definition,
        }
    }

    pub fn definition(&self) -> &SourceDefinition {
        &self.definition
    }
}

impl Display for VariableSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "VariableSymbol(name = {}, is_mutable = {})",
            self.name, self.is_mutable
        )
    }
}

#[derive(Debug, Clone)]
pub struct ModuleSymbol {
    pub fqn: String,
    pub declared_name: String,
    pub is_public: bool,
    definition: SourceDefinition,
}

impl ModuleSymbol {
    pub fn new(
        fqn: &str,
        declared_name: &str,
        is_public: bool,
        identifier: &Identifier,
    ) -> ModuleSymbol {
        ModuleSymbol {
            fqn: fqn.to_string(),
            declared_name: declared_name.to_string(),
            is_public,
            definition: SourceDefinition::from_identifier(identifier),
        }
    }

    pub fn definition(&self) -> &SourceDefinition {
        &self.definition
    }
}

impl Display for ModuleSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "ModuleSymbol(name = {}, is_public = {})",
            self.fqn, self.is_public
        )
    }
}
