use crate::vm::object::DvmObject;
use crate::vm::value::DvmValue;
use std::cell::RefCell;
use std::fmt::{Display, Write};
use std::rc::Rc;
use std::slice::Iter;

#[derive(Debug, PartialEq)]
pub enum DvmArray {
    Bytes(Vec<u8>),
    Ints(Vec<i32>),
    Longs(Vec<i64>),
    Doubles(Vec<f64>),
    USizes(Vec<usize>),
    Booleans(Vec<bool>),
    Objects(Vec<Rc<RefCell<DvmObject>>>),
    Arrays(Vec<Rc<RefCell<DvmArray>>>),
    Strings(Vec<Rc<String>>),
}

impl DvmArray {
    pub fn read_element(&self, index: usize) -> DvmValue {
        match self {
            DvmArray::Bytes(vec) => DvmValue::Byte(vec[index]),
            DvmArray::Ints(vec) => DvmValue::Int(vec[index]),
            DvmArray::Longs(vec) => DvmValue::Long(vec[index]),
            DvmArray::Doubles(vec) => DvmValue::Double(vec[index]),
            DvmArray::USizes(vec) => DvmValue::USize(vec[index]),
            DvmArray::Booleans(vec) => DvmValue::Boolean(vec[index]),
            DvmArray::Objects(vec) => DvmValue::Object(vec[index].clone()),
            DvmArray::Arrays(vec) => DvmValue::Array(vec[index].clone()),
            DvmArray::Strings(vec) => DvmValue::String(vec[index].clone()),
        }
    }

    pub fn write_element(&mut self, index: usize, value: DvmValue) {
        match self {
            DvmArray::Bytes(vec) => {
                vec[index] = value.expect_byte();
            }
            DvmArray::Ints(vec) => {
                vec[index] = value.expect_int();
            }
            DvmArray::Longs(vec) => {
                vec[index] = value.expect_long();
            }
            DvmArray::Doubles(vec) => {
                vec[index] = value.expect_double();
            }
            DvmArray::USizes(vec) => {
                vec[index] = value.expect_usize();
            }
            DvmArray::Booleans(vec) => {
                vec[index] = value.expect_boolean();
            }
            DvmArray::Objects(vec) => {
                vec[index] = value.expect_object();
            }
            DvmArray::Arrays(vec) => {
                vec[index] = value.expect_array();
            }
            DvmArray::Strings(vec) => {
                vec[index] = value.expect_string();
            }
        }
    }
}

impl Display for DvmArray {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "[{}]",
            self.into_iter()
                .map(|v| v.to_string())
                .collect::<Vec<_>>()
                .join(", ")
        )
    }
}

pub enum DvmArrayIter<'a> {
    Bytes(Iter<'a, u8>),
    Ints(Iter<'a, i32>),
    Longs(Iter<'a, i64>),
    Doubles(Iter<'a, f64>),
    USizes(Iter<'a, usize>),
    Booleans(Iter<'a, bool>),
    Objects(Iter<'a, Rc<RefCell<DvmObject>>>),
    Arrays(Iter<'a, Rc<RefCell<DvmArray>>>),
    Strings(Iter<'a, Rc<String>>),
}

impl<'a> Iterator for DvmArrayIter<'a> {
    type Item = DvmValue;
    fn next(&mut self) -> Option<Self::Item> {
        match self {
            DvmArrayIter::Bytes(iter) => iter.next().map(|b| DvmValue::Byte(*b)),
            DvmArrayIter::Ints(iter) => iter.next().map(|i| DvmValue::Int(*i)),
            DvmArrayIter::Longs(iter) => iter.next().map(|l| DvmValue::Long(*l)),
            DvmArrayIter::Doubles(iter) => iter.next().map(|d| DvmValue::Double(*d)),
            DvmArrayIter::USizes(iter) => iter.next().map(|u| DvmValue::USize(*u)),
            DvmArrayIter::Booleans(iter) => iter.next().map(|b| DvmValue::Boolean(*b)),
            DvmArrayIter::Objects(iter) => iter.next().map(|o| DvmValue::Object(o.clone())),
            DvmArrayIter::Arrays(iter) => iter.next().map(|a| DvmValue::Array(a.clone())),
            DvmArrayIter::Strings(iter) => iter.next().map(|s| DvmValue::String(s.clone())),
        }
    }
}

impl<'a> IntoIterator for &'a DvmArray {
    type Item = DvmValue;
    type IntoIter = DvmArrayIter<'a>;

    fn into_iter(self) -> Self::IntoIter {
        match self {
            DvmArray::Bytes(v) => DvmArrayIter::Bytes(v.iter()),
            DvmArray::Ints(v) => DvmArrayIter::Ints(v.iter()),
            DvmArray::Longs(v) => DvmArrayIter::Longs(v.iter()),
            DvmArray::Doubles(v) => DvmArrayIter::Doubles(v.iter()),
            DvmArray::USizes(v) => DvmArrayIter::USizes(v.iter()),
            DvmArray::Booleans(v) => DvmArrayIter::Booleans(v.iter()),
            DvmArray::Objects(v) => DvmArrayIter::Objects(v.iter()),
            DvmArray::Arrays(v) => DvmArrayIter::Arrays(v.iter()),
            DvmArray::Strings(v) => DvmArrayIter::Strings(v.iter()),
        }
    }
}
